#ifndef SINGLY_LINKED_LIST_H
#define SINGLY_LINKED_LIST_H

#include <cassert>
#include <string>
#include <type_traits>
#include "sequence.h"

namespace seq
{
   template<typename T> class SinglyLinkedList;
   
   template<typename T> class SinglyLinkedListNode
   {
   public:
      // return a reference to the stored item
      T& get_item() { return this->item; }

      // return pointer to next node, or nullptr if this is the final node
      SinglyLinkedListNode<T>* get_next() const { return this->next; }

      // overwrite the item stored in this node
      void set_item(T in_item) { this->item = in_item; }

   private:
      T item;
      SinglyLinkedListNode<T>* next = nullptr;

      // attach a node to this node
      // if there was a node attached to this previously, it is NOT deleted!
      void set_next(SinglyLinkedListNode<T>* in_next) { this->next = in_next; }
      
      friend class SinglyLinkedList<T>;  // allow SinglyLinkedList<T> to access private members
   };

   template<typename T> class SinglyLinkedList: public Sequence<T>
   {
   public:
      bool empty() const { return (this->head == nullptr); }  // test whether the singly linked list is empty
      size_t size() const;  // return the size (number of items in the singly linked list)

      // it is the caller's responsibility to ensure that the list is not empty when calling front() or back()!
      T& front() { assert(this->head); return this->head->get_item(); }  // return a reference to the first item
      T& back()  { assert(this->tail); return this->tail->get_item(); }  // return a reference to the final item
      
      // return a reference to the item at position i of the list, counting from 0
      // it is the caller's responsibility that the index is within range
      T& at(int i) { return this->index(i)->get_item(); }
      
      // return pointer to the head/tail node
      SinglyLinkedListNode<T>* begin() const { return this->head; }
      SinglyLinkedListNode<T>* end() const { return this->tail; }
      
      // return pointer to the node at position i, counting from 0
      SinglyLinkedListNode<T>* index(int i) const;

      /*
       * accepts an additional item into the singly linked list;
       * by default, this is done at the back end of the list
       * call push_front(...) to push an element at the front
       * 
       * the list takes ownership of the copy (but not of the original!)
       */
      void push(const T& pushed_item) { this->push_back(pushed_item); }
      void push_back(const T& pushed_item);
      void push_front(const T& pushed_item);

      /*
       * removes an item from the list (front end by default)
       * to do the same at the back, call pop_back()
       */
      void pop() { this->pop_front(); }
      void pop_front();
      void pop_back();
      void clear() { while(!this->empty()) this->pop(); }  // remove all the items from the list

      // it is the caller's responsibility that the index is within range
      void insert_at(int i, const T& inserted_item);  // insert an item at index i
      void erase_at(int i);  // remove the item at index i
      
      // it is the caller's responsibility that the node is actually part of the list
      void insert_successor_to(SinglyLinkedListNode<T>* predecessor, const T& inserted_item);  // insert an item after given node
      void erase_successor_to(SinglyLinkedListNode<T>* predecessor);  // remove the item after given node
      
      ~SinglyLinkedList() { this->clear(); }

   private:
      SinglyLinkedListNode<T>* head = nullptr;
      SinglyLinkedListNode<T>* tail = nullptr;
   };
}

#endif
